(** quations du premier ordre et leur rsolution *)

open T (** Pcf typ *)

(** {6 Pour le TP numro 6 } *)

type eqs =  (Type.t * Type.t) list
type subst
(** Types des ensembles d'quations et des substitutions *)

val print_eqs : out_channel -> eqs -> unit
val print_subst : out_channel -> subst -> unit
(** Et leurs afficheurs *)

val fresh_var : unit -> int
(** Renvoie une variable de type frache. *)

val subst : subst -> int -> Type.t
(** Applique une substitution  une variable de type *)

val subst_type : subst -> Type.t -> Type.t
(** L'appel [subst s t] applique la substitution [s] au type [t]. *)

val mgu : eqs -> subst
(**  Rsolution des quations, renvoie une solution principale. *)

val mgu_incr : Type.t -> Type.t -> subst -> subst
(** Rsolution incrmentale, on peut voir l'appel [mgu_incr t1 t2 s]
    comme renvoyant la substitution {i mgu(s(t1),s(t2)) o s}. *)

(** {6 Pour le TP numro 7 } *)

type scheme = int list * Type.t
(** Schma de type : une paire [(vs,t)], o [vs] est la liste des variables
    quantifies dans le type [ty]. *)

val print_scheme : out_channel -> scheme -> unit
(** Afficheur *)

(** Ensembles de variables de types (ensembles de [int], donc).
    Comme les ensembles de la bibliothque standard,
    avec deux fonctions supplmentaires.
@see <http://caml.inria.fr/pub/docs/manual-ocaml/libref/Set.S.html> Les ensembles de la bibliothque [Set] standard. *)
module Vars : sig

  include Set.S

  val of_list : elt list -> t
(** Fabrique un ensemble  partir d'une liste d'lments
    non ncessairement distincts deux  deux. *)

  val disjoint : t -> t -> bool
(** Teste si deux ensembles sont disjoints ({i i.e.} d'intersection vide). *)
end with type elt = int

val dom : subst -> Vars.t
(** Support d'une substitution, {i i.e.} ensemble des variables de type
    {i x} telles que {i s(x)} est diffrent de {i x}. *)

val id : subst
(** Substitution identite {i id(x) = x} pour tout {i x}. *)

val delta : int -> Type.t -> subst
(** L'appel [delta x ty] fabrique la substitution {i s(x) = ty}
    et {i s(y) = y} pour tout {i y} diffrent de {i x} *)

val comp : subst -> subst -> subst
(** Composition des substitutions,
    ne fonctionne que pour des substitutions de supports disjoints. *)

