open S
open Ast
open Code
open Printf

exception Error

(* Accu %eax, Stack %esp, Env %ebp *)
let eax = "%eax"
let ebx = "%ebx"
let ecx = "%ecx"
and ebp = "%ebp"
and esp = "%esp"
and esi = "%esi"

let car chan s = fprintf chan "    movl 0(%s),%%eax\n" s
and cdr chan s = fprintf chan "    movl 4(%s),%%eax\n" s
and push chan s = fprintf chan "    pushl %s\n" s
and pop chan s = fprintf chan "    popl %s\n" s
and move chan s d = fprintf chan "    movl %s,%s\n" s d

let alloc chan a1 a2 =
  push chan a2 ;
  push chan a1 ;
  fprintf chan "    call alloc\n" ;
  fprintf chan "    addl $8,%s\n" esp ;
  ()

let c_lab = ref 0

let gen_label () =
  let r = !c_lab in
  incr c_lab ;
  r

let pushenv chan = push chan ebp
and popenv chan =  pop chan ebp

let all_popenv c = List.for_all (fun i -> i = Popenv) c

let rec dump chan i k =
  fprintf chan "# %a\n" print_instruction i ;
  match i with
  | Ldi i ->
      move chan (sprintf "$%i" i) eax ; k
  | Push  ->
      push chan eax ; k
  | IOp Op.Add ->
      fprintf chan "    addl %%eax,0(%%esp)\n" ;
      pop chan eax ;
      k
  | IOp Op.Sub ->
      fprintf chan "    subl %%eax,0(%%esp)\n" ;
      pop chan eax ;
      move chan eax ecx ;
      fprintf chan "    notl %s\n" ecx ;
      fprintf chan "    sarl $31,%s\n" ecx ;
      fprintf chan "    andl %s,%s\n" ecx eax ;
      k
  | IOp Op.Mul ->
      move chan eax ecx ;
      pop chan eax ;
      fprintf chan "    imull %%ecx,%%eax\n" ;
      k
  | IOp Op.Div ->
      move chan eax ecx ;
      pop chan eax ;
      fprintf chan "    cltd\n" ;
      fprintf chan "    idivl %%ecx\n" ;
      k
  | Extend ->
      alloc chan eax ebp ;
      move chan eax ebp ; k
  | Search 0 -> car chan ebp ; k
  | Search n ->
      cdr chan ebp ;
      let rec search k = match k with
      | 0 -> car chan eax
      | _ -> cdr chan eax ; search (k-1)  in
      search (n-1) ;
      k
(* On peut ici tre un peu plus efficace,
   car on connat l'env dpil  l'avance. *)
  | Pushenv -> (* pushenv chan ; *) k
  | Popenv ->
(*      popenv chan ; *)
      move chan "4(%ebp)" ebp ;
      k
  | Mkclos c ->
      let lab = gen_label () in
      alloc chan (sprintf "$.L%i" lab) ebp ;
      (lab,c)::k
  | Mkclosrec c ->
      let lab = gen_label () in
      alloc chan eax ebp ;
      fprintf chan "    pushl %s\n" eax ;
      push chan (sprintf "$.L%i" lab) ;
      fprintf chan "    call alloc\n" ;
      fprintf chan "    addl $4,%s\n" esp ;
      fprintf chan "    popl %s\n" ecx ;
      fprintf chan "    movl %s,0(%s)\n" eax ecx ;
      (lab,c)::k
  | Apply ->
      pop chan ebx ;
      fprintf chan "    movl 0(%s),%s\n" eax ecx ;
      fprintf chan "    call *%s\n" ecx ;
      k
  | Test (_,_) -> assert false 

and dumps is_fun chan c k = match c with
| [] -> k
| Test (c2,c3)::c ->
    let lab1 = gen_label ()
    and lab2 = gen_label () in
    fprintf chan "#Test\n" ;
    fprintf chan "    testl %s,%s\n" eax eax ;
    fprintf chan "    je .L%i\n" lab2 ;
    let k = dumps (is_fun && all_popenv c) chan c3 k in
    fprintf chan "    jmp .L%i\n" lab1 ;
    fprintf chan  ".L%i:\n" lab2 ;
    let k = dumps  (is_fun && all_popenv c) chan c2 k in
    fprintf chan  ".L%i:\n" lab1 ;
    dumps is_fun chan c k
| Popenv::rem  when is_fun && all_popenv rem ->
    k
| Apply::rem when is_fun && all_popenv rem ->
    fprintf chan "#Tail%s\n"
      (let len = List.length rem in
      if len = 0 then ""
      else sprintf " + %i Popenv" len) ;
    pop chan ebx ;
    popenv chan ;
    fprintf chan "    movl 0(%s),%s\n" eax ecx ;
    fprintf chan "    jmp *%s\n" ecx ;
    k

| i::c ->
    let k = dump chan i k in
    dumps is_fun chan c k

and dumps_fun chan k = match k with
| [] -> ()
| (lab,is)::k ->
    fprintf chan ".L%i:\n" lab ;
    fprintf chan "#ENTER\n" ;
    pushenv chan ;
    fprintf chan "    movl 4(%s),%s\n" eax eax ;
    alloc chan ebx eax ;
    move chan eax ebp ;
    let k = dumps true chan is k in
    fprintf chan "#RET\n" ;
    popenv chan ;
    fprintf chan "    ret\n" ;
    dumps_fun chan k

let callee_saves = [ebp; ebx; esi]

let top_dumps chan is =
  fprintf chan "#START\n" ;
  fprintf chan "    .globl pcf\n" ;
  fprintf chan "pcf:\n" ;
  List.iter (push chan) callee_saves ;
  fprintf chan "    xorl %s,%s\n" ebp ebp ;
  let k = dumps false chan is [] in
  fprintf chan "#FINISH\n" ;
  List.iter (pop chan) (List.rev callee_saves) ;
  fprintf chan "    ret\n" ;
  dumps_fun chan k ;
  ()

let finally f a k =
  try let r = f a in k () ; r
  with e -> k () ; raise e

let dump_code chan is =
  finally
    (top_dumps chan) is
    (fun () -> ())

let compile chan t = dump_code chan(Compiler.compile [] t)
