%{
open Op
open Ast
%}
%token <int> INT
%token <string> IDENT
%token ADD SUB MUL DIV
%token EQUAL ARROW
%token OPAR CPAR
%token FUN FIX
%token LET IN REC
%token IFZ THEN ELSE
%token EOF SEMISEMI
%token COLON NAT
%token OA CA
%token DOT
%token WITH
%token OBJ
%token SHARP
%token SEMI

/* lowest precedence */
%nonassoc ARROW
%nonassoc IN
%nonassoc ELSE
%left  ADD SUB
%left  MUL DIV


/* Untyped PCF */
%start main             
%type <Ast.t> main
%start top             
%type <string option * Ast.t> top

%%

main:
| t3 EOF { $1 }
| t3 SEMISEMI EOF { $1 }

top:
| t3 SEMISEMI { (None,$1) }
| LET IDENT EQUAL t3 SEMISEMI { (Some $2,$4) }
| LET REC IDENT EQUAL t3 SEMISEMI { (Some $3,Fix ($3,$5)) }
| EOF { raise Exit }

t3:
| t2   { $1 }
| FUN IDENT ARROW t3 { Fun ($2,$4) }
| FIX IDENT ARROW t3 { Fix ($2,$4) }
| LET IDENT EQUAL t3 IN t3 { Let ($2,$4,$6) }
| LET REC IDENT EQUAL t3 IN t3 { Let ($3,Fix ($3,$5),$7) }
| IFZ t3 THEN t3 ELSE t3 { Ifz ($2, $4,$6) }


t2:
| t1 { $1 }
| t2 ADD t2 { Op (Add,$1,$3) }
| t2 SUB t2 { Op (Sub,$1,$3) }
| t2 MUL t2 { Op (Mul,$1,$3) }
| t2 DIV t2 { Op (Div,$1,$3) } 
 
t1:
| t0 { $1 }
| t1 t0 { App($1, $2) }

t0:
| INT { Num $1 }
| IDENT { Var $1 }
| OPAR t3 CPAR { $2 }

| OA labels CA { Rec $2 }
| OA t3 WITH labels CA { CopyWith ($2,$4) }
| t0 DOT IDENT { Access ($1,$3) }

| OBJ OA labels CA { Obj (None,$3) }
| OBJ IDENT OA labels CA { Obj (Some $2, $4) }
| OBJ OA t3 WITH labels CA { ObjWith (None, $3, $5) }
| OBJ IDENT OA t3 WITH labels CA { ObjWith (Some $2, $4, $6) }
| t0 SHARP IDENT { Invoke ($1,$3) }

labels:
| { [ ] }
| IDENT EQUAL t3 SEMI labels { ($1,$3) :: $5 }
| IDENT EQUAL t3 { [$1,$3] }
