type t =
  | Num of int              (** Constante entire *)
  | Var of string           (** Accs  une variable *)
  | Op  of Op.t * t * t     (** Application d'un oprateur binaire [t1 op t2] *)
  | Ifz of t * t * t        (** Conditionnelle [Ifz t1 Then t2 Else t3] *)
  | Let of string * t * t   (** Liaison locale [Let x = t1 In t2] *) 
  | App of t * t            (** Application de fonction [t1 t2]*)
  | Fun of string * t       (** Dfinition de fonction [Fun x -> t] *)
  | Fix of string * t       (** Point fixe [Fix x -> t] *)

open Printf

let rec p_term p_bind chan = function
  | Num n -> fprintf chan "%i" n
  | Var x -> fprintf chan "%s" x
  | Op (op,t1,t2) ->
      fprintf chan "%a%c%a"
        (p_arg p_bind) t1
        (Op.to_char op)
        (p_arg p_bind) t2
  | App (t1,t2) ->
      fprintf chan "%a %a" (p_fun p_bind) t1 (p_arg p_bind) t2
  | Fun (x,t1) ->
      fprintf chan "Fun %a -> %a" p_bind x (p_term p_bind) t1
  | Fix (f,t1) ->
      fprintf chan "Fix %a -> %a" p_bind f (p_term p_bind) t1
  | Let (x,tx,t) ->
      fprintf chan "Let %s = %a In %a"
        x (p_term p_bind) tx (p_term p_bind) t
  | Ifz (t1,t2,t3) ->
      fprintf chan "Ifz %a Then %a Else %a"
        (p_term p_bind) t1 (p_term p_bind) t2 (p_term p_bind) t3

and p_arg p_bind chan t = match t with
  | Num _ | Var _ ->
      (p_term p_bind) chan t
  | _ -> fprintf chan "(%a)" (p_term p_bind) t

and p_fun p_bind chan t = match t with
  | Num _ | Var _ | App (_,_) -> (p_term p_bind) chan t
  | _ -> fprintf chan "(%a)" (p_term p_bind) t

let print chan t = p_term output_string chan t

